################################################################################################################################

#	Citizens United, States Divided
#	Douglas M. Spencer and Abby K. Wood

#	This Python code scrapes the website of the National Institute for Money in State Politics (www.followthemoney.org)
#	for detailed information about the source and target of independent expenditures in state elections.

#	Adapted from code created by Aamir Adnan -- PART I

#	Code last updated March 3, 2012

################################################################################################################################

import urllib2
import urllib
import cookielib
import re, random
import os, sys, inspect
import xlrd
import xlwt
import time

################################################################################################################################

# Set working diretory
os.chdir("/Users/Doug/Dropbox/Python/")

# regexes used in extraction process
count_reg = re.compile('<div class="Transaction">', re.IGNORECASE)
sector_reg = re.compile('<div class="reportbox">\s*<img[^>]+>\s*<p><span class="bold">([^<]*)</span><BR>([^<]*)<BR><small>([^<]*)</small></p></div>', re.IGNORECASE)

target_intermediate_reg = re.compile('<TABLE class="DBtable">\s*<caption>TABLE \d+: #(\d+) Targets</caption>\s*<thead><tr><th>Position</th>\s*<th>Target</th>\s*<th>Target Info</th>\s*<th>Targeted Amount</th>\s*</tr></thead>([\s\S]+?)</table>', re.IGNORECASE)
payees_intermediate_reg = re.compile('<TABLE class="DBtable">\s*<caption>TABLE \d+: #(\d+) Payees</caption>\s*<thead><tr><th>Payee</th>\s*<th>Purpose</th>\s*<th>Type</th>\s*<th>Date</th>\s*<th>Amount</th>\s*</tr></thead>([\s\S]+?)</table>', re.IGNORECASE)
target_reg = re.compile('<tr class="grey\d+"><td class="text-center">([^<]+)</td>\s*<td class="text-left">(?:<a[^>]+>)?([^<]+)(?:</a>)?\s*</td>\s*<td class="small text-center">([^<]+)</td>\s*<td class="text-right">([^<]+)</td>\s*</tr>', re.IGNORECASE)
payees_reg = re.compile('<tr class="grey\d+"><td[^>]*>([^<]*)</td>\s*<td class="small text-center">([^<]*)</td>\s*(?:<td class="small text-center">([^<]+)</td>)?\s*<td class="text-center">([^<]+)</td>\s*<td class="text-right">([^<]+)</td>\s*</tr>', re.IGNORECASE)

# Defining headers for csv files
IEsummary1_header = ['State', 'Year', 'Spender', 'Num.Races', 'Num.Ballot', 'Num.Cmtes', 'Num.Expend', 'Total.Spent', 'Sector1', 'Sector2', 'Sector3', 'Hyperlink']
IEdetail1_header = ['State', 'Spender', 'Sector1', 'Sector2', 'Sector3', 'Position', 'Target', 'Seat', 'Party', 'Amount', 'Date', 'Expenditure Num']
IEpayees1_header = ['State', 'Spender', 'Sector1', 'Sector2', 'Sector3', 'Expenditure Num', 'Payee', 'Purpose', 'Type', 'Date', 'Amount']

# File which is used to read hyperlinks
mainData_book = xlrd.open_workbook("IEsummary.xls", formatting_info=True)
mainData_sheet = mainData_book.sheet_by_index(0) # Get the first sheet 0

# Create new csv for IEsummary
IEsummary1 = xlwt.Workbook()
IEsummary1_sheet = IEsummary1.add_sheet('IEsummary')
for i in range(len(IEsummary1_header)):
    IEsummary1_sheet.write(0, i, IEsummary1_header[i])
    
# Create new csv for IEsummary Invalid
IEsummary1_invalid = xlwt.Workbook()
IEsummary1_invalid_sheet = IEsummary1_invalid.add_sheet('IEsummary_invalid')
for i in range(len(IEsummary1_header)):
    IEsummary1_invalid_sheet.write(0, i, IEsummary1_header[i])

# Create new csv for IEdetail
IEdetail1 = xlwt.Workbook()
IEdetail1_sheet = IEdetail1.add_sheet('IEdetail')
for i in range(len(IEdetail1_header)):
    IEdetail1_sheet.write(0, i, IEdetail1_header[i])

# Create new csv for Payees
IEpayees1 = xlwt.Workbook()
IEpayees1_sheet = IEpayees1.add_sheet('IEpayees')
for i in range(len(IEpayees1_header)):
    IEpayees1_sheet.write(0, i, IEpayees1_header[i])

def saveError(html):
    file = open('error.html', 'w')
    file.write(html)
    file.close()

def getContent(cOpener, url, post=None, headers={}):
    print url
    if post:
        post = urllib.urlencode(post)
    req = urllib2.Request(url, post, headers)
    data   = cOpener.open(req)
    data = data.read()
    nm = random.random()
    if not os.path.exists('response'):
        os.makedirs('response')
    #file = open("response/html"+str(nm)+".html", "w")
    #file.write("######################\nURL:"+url+"\nPOST:"+str(post)+"\n######################\n\n"+data)
    #file.close()
    return data

def getOpener():
    opnr = urllib2.build_opener(urllib2.HTTPCookieProcessor(cookielib.CookieJar()))
    urllib2.install_opener( opnr )
    return opnr
    
def extraction(html, ieInfo):
    print "Doing extraction............."   
    global rowCountForIEdetail1
    global rowCountForIEpayee1
    payeeFirstDate = None
    total = count_reg.findall(html)
    targets = target_intermediate_reg.findall(html)
    payees = payees_intermediate_reg.findall(html)
    
    if (len(total) == 0 and len(targets) == 0) or (len(total) != len(targets)):
        saveError(html)
        raise 'Expected %s number of matches for targets found %s, please check error page.' % (len(total), len(targets))
        
    if (len(total) == 0 and len(payees) == 0) or (len(total) != len(payees)):
        saveError(html)
        raise 'Expected %s number of matches for payees found %s, please check error page.' % (len(total), len(targets))

    for multiplePayees in payees:
        expeditureNum = multiplePayees[0]
        if payees_reg.search(multiplePayees[1]):
            payees_info = payees_reg.findall(multiplePayees[1])
            for eachPayee in payees_info:
                payee = eachPayee[0].strip()
                purpose = eachPayee[1]
                type = eachPayee[2]
                date = eachPayee[3].strip()
                if not payeeFirstDate:
                    payeeFirstDate = date
                amount = eachPayee[4].strip()
                
                payeeTuple = (ieInfo[0], ieInfo[1], ieInfo[2], ieInfo[3], ieInfo[4], expeditureNum, payee, purpose, type, date, amount)
                saveToExcel(payeeTuple, rowCountForIEpayee1, IEpayees1_sheet)
                rowCountForIEpayee1 += 1
        else:
            saveError(html)
            raise 'Payees regex is not working, please check error page.'
        
    for multipleTargets in targets:
        expeditureNum = multipleTargets[0]
        if target_reg.search(multipleTargets[1]):
            records = target_reg.findall(multipleTargets[1])
            for eachRecord in records:
                positon = eachRecord[0].strip()
                target = eachRecord[1].strip()
                if '|' in eachRecord[2]:
                    temp = eachRecord[2].strip().split('|')
                    seat = temp[0].strip()
                    party = temp[1].strip()
                else:
                    seat = eachRecord[2].strip()
                    party = ''
                amount = eachRecord[3].strip()
                dataTuple = (ieInfo[0], ieInfo[1], ieInfo[2], ieInfo[3], ieInfo[4], positon, target, seat, party, amount, payeeFirstDate, expeditureNum)
                saveToExcel(dataTuple, rowCountForIEdetail1, IEdetail1_sheet)
                rowCountForIEdetail1 += 1
        else:
            saveError(html)
            raise 'Targets regex is not working, please check error page.'
    print "Extraction Completed.............\n"

def saveToExcel(dataTuple, row, sheet):
    for i in range(len(dataTuple)):
        sheet.write(row, i, dataTuple[i])

if __name__ == "__main__":
    rowCountForIEdetail1 = 1
    rowCountForIEpayee1 = 1
    rowCountForIEinvalid1 = 1
    _web = getOpener()
    for counter in range(1, 2000):
        print "Processing %s of %s." % (counter, 2000)
        rowValues = mainData_sheet.row_values(counter, start_colx=0, end_colx=8)
        link = mainData_sheet.hyperlink_map.get((counter, 2))
        link = link.url_or_path
        html = getContent(_web, link)
        
        invalid = re.search('(?i)<p class="text-center">There were no results found to display\.</p>', html)

        if invalid:
            dataTuple = (rowValues[0], rowValues[1], rowValues[2], rowValues[3], rowValues[4], rowValues[5], rowValues[6], rowValues[7])
            saveToExcel(dataTuple, rowCountForIEinvalid1, IEsummary1_invalid_sheet)
            rowCountForIEinvalid1 +=1
        else:
            if re.search('(?i)<TITLE>Access Limit Exceeded</TITLE>', html):
                print 'Sleeping for 60 seconds.. Access Limit Exceeded'
                time.sleep(60)
                html = getContent(_web, link)
            if sector_reg.search(html):
                sector_info = sector_reg.search(html)
                sector1 = sector_info.group(1)
                sector2 = sector_info.group(2)
                sector3 = sector_info.group(3)
            else:
                saveError(html)
                raise 'Sector info not found, please check error page.'
                
            state = rowValues[0]
            spender = rowValues[2]
            
            dataTuple = (rowValues[0], rowValues[1], rowValues[2], rowValues[3], rowValues[4], rowValues[5], rowValues[6], rowValues[7], sector1, sector2, sector3, link)
            saveToExcel(dataTuple, counter, IEsummary1_sheet)
            extraction(html, (state, spender, sector1, sector2, sector3))
    IEsummary1.save('IEsummary1.csv')
    IEdetail1.save('IEdetail1.csv')
    IEpayees1.save('IEpayees1.csv')
    IEsummary1_invalid.save('IEinvalid1.csv')
    print 'process completed successfully..'

####################################
######### Now run NIMSP2 ###########
####################################
